//
// GameX - Input-related GameXClass Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


#include "gamex-win-dx.h"


inline long long GetClocksPerSec();

bool GameXClass::z_UpdateInput ()
{
	// give DirectInput time to catch its breath
//	SwitchToThread();
///	Sleep(1); z_sleep_record += GetClocksPerSec() / CLOCKS_PER_SEC;

	bool skip_keyboard = (z_di_keyboard == NULL);
	bool skip_mouse = (z_di_mouse == NULL) || z_win_request_no_mouse_input;

	if(!skip_keyboard)
	{
		// get current physical state
		if(FAILED(z_di_keyboard->GetDeviceState(256, (LPVOID)z_di_keystate)))
		{
			if(SUCCEEDED(z_di_keyboard->Acquire()))
			{
				if(FAILED(z_di_keyboard->GetDeviceState(256, (LPVOID)z_di_keystate)))
				{
					skip_keyboard = true;
				}
			}
		}

		if(!skip_keyboard)
		{
			{
				for(KeyID k = 0 ; k < 255 ; k++)
				{
					if(z_di_keypressed[k] & INPUT_PRESSED) // if just pressed last time,
						z_di_keypressed[k] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}

			// fix windows oddity that makes the right shift button register as unpressed
			// if an arrow key is pressed while it's down
			if(!IsKeyDown(KEY_RSHIFT) && (GetAsyncKeyState(VK_RSHIFT) & 0x8000))
				z_di_keystate[KEY_RSHIFT] |= INPUT_PRESSED;

			// add in buffered updates
			{
				DIDEVICEOBJECTDATA di_keybuf [DX_KEY_BUFSIZE];
				DWORD elements = DX_KEY_BUFSIZE;
				if(SUCCEEDED(z_di_keyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA), di_keybuf, &elements, 0)))
				{
					if(!z_di_mousestop) // throw it out if just switched from another application, otherwise:
					{
						static std::stack<KeyID> newKeyUps; // intentionally a stack to re-reverse order
						static std::stack<KeyID> newKeyDowns;
						for(int i = (int)elements-1 ; i >= 0 ; i--) // reverse order so press+release -> press registered on this frame
						{
							if(di_keybuf[i].dwData & INPUT_PRESSED)
							{
								// it was pressed
								z_di_keystate[di_keybuf[i].dwOfs] = INPUT_PRESSED;
								z_di_keypressed[di_keybuf[i].dwOfs] = INPUT_PRESSED;
								newKeyDowns.push(di_keybuf[i].dwOfs);
							}
							else
							{
								// it was released
								z_di_keystate[di_keybuf[i].dwOfs] = 0;
								newKeyUps.push(di_keybuf[i].dwOfs);
							}
						}
						while(!newKeyDowns.empty())
							z_win_keydownbuf.push(newKeyDowns.top()), newKeyDowns.pop();
						while(!newKeyUps.empty())
							z_win_keyupbuf.push(newKeyUps.top()), newKeyUps.pop();
					}
				}
			}

			// transfer/record unique (non-held) keypresses
			for(KeyID k = 0 ; k < 255 ; k++)
			{
				if(z_di_keypressed[k] & INPUT_HELD)
				{
					if(!z_di_keystate[k])
						z_di_keypressed[k] = z_di_keystate[k];
				}
				else
				{
					z_di_keypressed[k] = z_di_keystate[k];
				}
			}

			if(z_IsEscapeReserved() && GetKeyPress(KEY_ESCAPE) && !(IsShiftDown() && IsCtrlDown()) // ESC is reserved to quit unless RUN_USEESCAPEKEY is on
			|| IsCtrlDown() && GetKeyPress(KEY_Q) // whenever CTRL is down, Q is reserved to quit (nonstandard but intuitive)
			|| IsAltDown() && GetKeyPress(KEY_F4)) // whenever ALT is down, F4 is reserved to quit (faster than in WindowFunc)
				Quit();


			if(z_IsAltEnterReserved() && GetKeyPress(KEY_ENTER) && IsAltDown()) // ALT+ENTER is reserved to toggle fullscreen mode unless RUN_NOALTENTERTOGGLE is on
				GameX.ToggleFullscreen();
		}
	}

	if(!skip_mouse)
	{

		if(FAILED(z_di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)z_di_mousestate)))
		{
			if(SUCCEEDED(z_di_mouse->Acquire()))
			{
				if(FAILED(z_di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)z_di_mousestate)))
				{
					skip_mouse = true;
				}
			}
		}

		if(!skip_mouse)
		{

			{
				for(MouseButtonID b = 0 ; b < 4 ; b++)
				{
					if(z_di_mousepressed[b] & INPUT_PRESSED) // if just pressed last time,
						z_di_mousepressed[b] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}
			
			// if the mouse isn't outside of the game,
			if(z_win_fullscreen || (z_mouse_x>0 && z_mouse_y>0 && GetMouseX()<z_win_width && GetMouseY()<z_win_height))
			{
				// add in buffered updates
				{
					DIDEVICEOBJECTDATA di_buttonbuf [DX_KEY_BUFSIZE];
					DWORD elements = DX_KEY_BUFSIZE;
					if(SUCCEEDED(z_di_mouse->GetDeviceData(sizeof(DIDEVICEOBJECTDATA), di_buttonbuf, &elements, 0)))
					{
						static std::stack<KeyID> newButtonUps; // intentionally a stack to re-reverse order
						static std::stack<KeyID> newButtonDowns;
						for(int i = (int)elements-1 ; i >= 0 ; i--) // reverse order so press+release -> press registered on this frame
						{
							int buttonPressed = -1;
							// Grouped ifs rather than switch because 
							// DIMOFS_* are not constant expressions to gcc
							if (di_buttonbuf[i].dwOfs == DIMOFS_BUTTON0) {
								buttonPressed = 0;
							} else if (di_buttonbuf[i].dwOfs == DIMOFS_BUTTON1) {
								buttonPressed = 1;
							} else if (di_buttonbuf[i].dwOfs == DIMOFS_BUTTON2) {
								buttonPressed = 2;
							} else if (di_buttonbuf[i].dwOfs == DIMOFS_BUTTON3) {
								buttonPressed = 3;
							}

							if(buttonPressed != -1)
							{
								if(di_buttonbuf[i].dwData & INPUT_PRESSED)
								{
									// it was pressed
									z_di_mousestate->rgbButtons[buttonPressed] = INPUT_PRESSED;
									z_di_mousepressed[buttonPressed] = INPUT_PRESSED;
									newButtonDowns.push(buttonPressed);
								}
								else
								{
									// it was released
									z_di_mousestate->rgbButtons[buttonPressed] = 0;
									newButtonUps.push(buttonPressed);
								}
							}
						}
						while(!newButtonDowns.empty())
							z_win_mousedownbuf.push(newButtonDowns.top()), newButtonDowns.pop();
						while(!newButtonUps.empty())
							z_win_mouseupbuf.push(newButtonUps.top()), newButtonUps.pop();
					}
				}

				// transfer/record unique (non-held) mousepresses
				for(MouseButtonID b = 0 ; b < 4 ; b++)
				{
					if(z_mouse_down[b] == 2 && z_di_mousestate->rgbButtons[b])
						z_di_mousestate->rgbButtons[b] |= INPUT_DOUBLE;
					if(z_di_mousepressed[b] & INPUT_HELD)
					{
						if(!z_di_mousestate->rgbButtons[b])
							z_di_mousepressed[b] = z_di_mousestate->rgbButtons[b];
					}
					else
					{
						z_di_mousepressed[b] = z_di_mousestate->rgbButtons[b];
					}
				}
			}
			if(z_di_mousestop)
			{
				z_di_mousestate->lX = z_di_mousestate->lY = z_di_mousestate->lZ = 0;
				if(z_win_active)
					z_di_mousestop--;

				// unregister all clicks at this time (so the game doesn't respond to window-activating clicks)
				for(MouseButtonID b = 0 ; b < 4 ; b++)
				{
					z_mouse_down[b] = 0;
					z_di_mousepressed[b] = 0;
				}
			}

			POINT mousePoint;
			GetCursorPos(&mousePoint);
			mousePoint.x -= z_win_xpos+z_win_xoffset;
			mousePoint.y -= z_win_ypos+z_win_yoffset;
			z_SetMouseMove(mousePoint.x,mousePoint.y);
		}
	}

	return true;
}

bool GameXClass::AnyKeyPress ()	
{
	for(int n = 0; n<256; n++)
		if(GetKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}

bool GameXClass::PeekAnyKeyPress ()	
{
	for(int n = 0; n<256; n++)
		if(PeekKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}


void GameXClass::GetMouseButtonString (char * output, MouseButtonID button)
{
	const char* result;
	switch(button)
	{
		case MOUSE_LEFT: result = "LClick"; break;
		case MOUSE_RIGHT: result = "RClick"; break;
		case MOUSE_MIDDLE: result = "MClick"; break;
		case MOUSE_EXTRA: result = "EClick"; break;
		default: result = "unknown"; break;
	}
	strcpy(output,result);
}



void GameXClass::GetKeyString (char * output, KeyID key)
{
	const char* result;
	switch(key)
	{
		case KEY_SPACE: result = "SpaceBar"; break;
		case KEY_ENTER: result = "Enter"; break;
		case KEY_A: result = "A"; break;
		case KEY_B: result = "B"; break;
		case KEY_C: result = "C"; break;
		case KEY_D: result = "D"; break;
		case KEY_E: result = "E"; break;
		case KEY_F: result = "F"; break;
		case KEY_G: result = "G"; break;
		case KEY_H: result = "H"; break;
		case KEY_I: result = "I"; break;
		case KEY_J: result = "J"; break;
		case KEY_K: result = "K"; break;
		case KEY_L: result = "L"; break;
		case KEY_M: result = "M"; break;
		case KEY_N: result = "N"; break;
		case KEY_O: result = "O"; break;
		case KEY_P: result = "P"; break;
		case KEY_Q: result = "Q"; break;
		case KEY_R: result = "R"; break;
		case KEY_S: result = "S"; break;
		case KEY_T: result = "T"; break;
		case KEY_U: result = "U"; break;
		case KEY_V: result = "V"; break;
		case KEY_W: result = "W"; break;
		case KEY_X: result = "X"; break;
		case KEY_Y: result = "Y"; break;
		case KEY_Z: result = "Z"; break;
		case KEY_0: result = "0"; break;
		case KEY_1: result = "1"; break;
		case KEY_2: result = "2"; break;
		case KEY_3: result = "3"; break;
		case KEY_4: result = "4"; break;
		case KEY_5: result = "5"; break;
		case KEY_6: result = "6"; break;
		case KEY_7: result = "7"; break;
		case KEY_8: result = "8"; break;
		case KEY_9: result = "9"; break;
		case KEY_TILDE: result = "`"; break;
		case KEY_APOSTROPHE: result = "'"; break;
		case KEY_SEMICOLON: result = ";"; break;
		case KEY_LBRACKET: result = "["; break;
		case KEY_RBRACKET: result = "]"; break;
		case KEY_COMMA: result = ","; break;
		case KEY_PERIOD: result = "."; break;
		case KEY_NUM0: result = "num0"; break;
		case KEY_NUM1: result = "num1"; break;
		case KEY_NUM2: result = "num2"; break;
		case KEY_NUM3: result = "num3"; break;
		case KEY_NUM4: result = "num4"; break;
		case KEY_NUM5: result = "num5"; break;
		case KEY_NUM6: result = "num6"; break;
		case KEY_NUM7: result = "num7"; break;
		case KEY_NUM8: result = "num8"; break;
		case KEY_NUM9: result = "num9"; break;
		case KEY_NUMTIMES: result = "num*"; break;
		case KEY_NUMDIVIDE: result = "num/"; break;
		case KEY_NUMPLUS: result = "num+"; break;
		case KEY_NUMMINUS: result = "num-"; break;
		case KEY_NUMDOT: result = "NUM."; break;
		case KEY_LEFT: result = "LeftArrow"; break;
		case KEY_UP: result = "UpArrow"; break;
		case KEY_RIGHT: result = "RightArrow"; break;
		case KEY_DOWN: result = "DownArrow"; break;
		case KEY_CAPS: result = "CapsLock"; break;
		case KEY_TAB: result = "Tab"; break;
		case KEY_ESCAPE: result = "Escape"; break;
		case KEY_LSHIFT: result = "LShift"; break;
		case KEY_RSHIFT: result = "RShift"; break;
		case KEY_LALT: result = "LAlt"; break;
		case KEY_RALT: result = "RAlt"; break;
		case KEY_LWIN: result = "LWin"; break;
		case KEY_RWIN: result = "RWin"; break;
		case KEY_LCTRL: result = "LCtrl"; break;
		case KEY_RCTRL: result = "RCtrl"; break;
		case KEY_BACKSLASH: result = "/"; break;
		case KEY_FORWARDSLASH: result = "\\"; break;
		case KEY_MINUS: result = "-"; break;
		case KEY_PLUS: result = "="; break;
		case KEY_BACKSPACE: result = "Backspace"; break;
		case KEY_DELETE: result = "Delete"; break;
		case KEY_HOME: result = "Home"; break;
		case KEY_END: result = "End"; break;
		case KEY_INSERT: result = "Insert"; break;
		case KEY_PAGEUP: result = "PageUp"; break;
		case KEY_PAGEDOWN: result = "PageDown"; break;
		case KEY_PAUSE: result = "Pause"; break;
		case KEY_NUMLOCK: result = "NumLock"; break;
		case KEY_PRINT: result = "PrintScreen"; break;
		case KEY_F1: result = "F1"; break;
		case KEY_F2: result = "F2"; break;
		case KEY_F3: result = "F3"; break;
		case KEY_F4: result = "F4"; break;
		case KEY_F5: result = "F5"; break;
		case KEY_F6: result = "F6"; break;
		case KEY_F7: result = "F7"; break;
		case KEY_F8: result = "F8"; break;
		case KEY_F9: result = "F9"; break;
		case KEY_F10: result = "F10"; break;
		case KEY_F11: result = "F11"; break;
		case KEY_F12: result = "F12"; break;
		default: result = "unknown"; break;
	}
	strcpy(output,result);
}




KeyID GameXClass::GetKeyDownEvent ()
{
	if(!z_win_keydownbuf.empty())
	{
		KeyID ret = z_win_keydownbuf.front();
		z_win_keydownbuf.pop();
		return ret;
	}
	return KEY_NOTHING;
}
KeyID GameXClass::GetKeyUpEvent ()
{
	if(!z_win_keyupbuf.empty())
	{
		KeyID ret = z_win_keyupbuf.front();
		z_win_keyupbuf.pop();
		return ret;
	}
	return KEY_NOTHING;
}

KeyID GameXClass::PeekKeyDownEvent ()
{
	if(!z_win_keydownbuf.empty())
		return z_win_keydownbuf.front();
	return KEY_NOTHING;
}

KeyID GameXClass::PeekKeyUpEvent ()
{
	if(!z_win_keyupbuf.empty())
		return z_win_keyupbuf.front();
	return KEY_NOTHING;
}

void GameXClass::ClearKeyDownEvents ()
{
	while(!z_win_keydownbuf.empty())
		z_win_keydownbuf.pop();
}

void GameXClass::ClearKeyUpEvents ()
{
	while(!z_win_keyupbuf.empty())
		z_win_keyupbuf.pop();
}

MouseButtonID GameXClass::GetMouseDownEvent ()
{
	if(!z_win_mousedownbuf.empty())
	{
		MouseButtonID ret = z_win_mousedownbuf.front();
		z_win_mousedownbuf.pop();
		return ret;
	}
	return MOUSE_NOTHING;
}
MouseButtonID GameXClass::GetMouseUpEvent ()
{
	if(!z_win_mouseupbuf.empty())
	{
		MouseButtonID ret = z_win_mouseupbuf.front();
		z_win_mouseupbuf.pop();
		return ret;
	}
	return MOUSE_NOTHING;
}
MouseButtonID GameXClass::PeekMouseDownEvent ()
{
	if(!z_win_mousedownbuf.empty())
		return z_win_mousedownbuf.front();
	return MOUSE_NOTHING;
}
MouseButtonID GameXClass::PeekMouseUpEvent ()
{
	if(!z_win_mouseupbuf.empty())
		return z_win_mouseupbuf.front();
	return MOUSE_NOTHING;
}
void GameXClass::ClearMouseDownEvents ()
{
	while(!z_win_mousedownbuf.empty())
		z_win_mousedownbuf.pop();
}
void GameXClass::ClearMouseUpEvents ()
{
	while(!z_win_mouseupbuf.empty())
		z_win_mouseupbuf.pop();
}

char GameXClass::GetBufferedKeyPress ()
{
	if(!z_win_keycharbuf.empty())
	{
		int ret = z_win_keycharbuf.front();
		z_win_keycharbuf.pop();
		return ret;
	}
	return '\0';
}

char GameXClass::PeekBufferedKeyPress ()
{
	if(!z_win_keycharbuf.empty())
		return z_win_keycharbuf.front();
	return '\0';
}

void GameXClass::ClearKeyBuffer ()
{
	while(!z_win_keycharbuf.empty())
		z_win_keycharbuf.pop();
}

void GameXClass::z_AddBufferedKeyPress (char ch)
{
	if(z_win_keycharbuf.size() < WIN_KEY_BUFSIZE)
		z_win_keycharbuf.push(ch);
}
void GameXClass::z_AddKeyDownEvent (KeyID key)
{
	if(z_win_keydownbuf.size() < WIN_KEY_BUFSIZE)
		z_win_keydownbuf.push(key);
}
void GameXClass::z_AddKeyUpEvent (KeyID key)
{
	if(z_win_keyupbuf.size() < WIN_KEY_BUFSIZE)
		z_win_keyupbuf.push(key);
}
void GameXClass::z_AddMouseDownEvent (MouseButtonID btn)
{
	if(z_win_mousedownbuf.size() < WIN_KEY_BUFSIZE)
		z_win_mousedownbuf.push(btn);
}
void GameXClass::z_AddMouseUpEvent (MouseButtonID btn)
{
	if(z_win_mouseupbuf.size() < WIN_KEY_BUFSIZE)
		z_win_mouseupbuf.push(btn);
}


